@@ -18,10 +18,11 @@
 #include <limits.h>
 #include <stdarg.h>
 
-#include "bson-compat.h"
-#include "bson-string.h"
-#include "bson-memory.h"
-#include "bson-utf8.h"
+#include <bson/bson-compat.h>
+#include <bson/bson-config.h>
+#include <bson/bson-string.h>
+#include <bson/bson-memory.h>
+#include <bson/bson-utf8.h>
 
 #ifdef BSON_HAVE_STRINGS_H
 #include <strings.h>
@@ -76,41 +77,21 @@ bson_string_new (const char *str) /* IN */
    if (str) {
       memcpy (ret->str, str, ret->len);
    }
-   ret->str[ret->len] = '\0';
 
    ret->str[ret->len] = '\0';
 
    return ret;
 }
 
-
-/*
- *--------------------------------------------------------------------------
- *
- * bson_string_free --
- *
- *       Free the bson_string_t @string and related allocations.
- *
- *       If @free_segment is false, then the strings buffer will be
- *       returned and is not freed. Otherwise, NULL is returned.
- *
- * Returns:
- *       The string->str if free_segment is false.
- *       Otherwise NULL.
- *
- * Side effects:
- *       None.
- *
- *--------------------------------------------------------------------------
- */
-
 char *
 bson_string_free (bson_string_t *string, /* IN */
                   bool free_segment)     /* IN */
 {
    char *ret = NULL;
 
-   BSON_ASSERT (string);
+   if (!string) {
+      return NULL;
+   }
 
    if (!free_segment) {
       ret = string->str;
@@ -492,11 +473,11 @@ bson_strndup (const char *str, /* IN */
 void
 bson_strfreev (char **str) /* IN */
 {
-   int i;
-
    if (str) {
-      for (i = 0; str[i]; i++)
-         bson_free (str[i]);
+      for (char **ptr = str; *ptr != NULL; ++ptr) {
+         bson_free (*ptr);
+      }
+
       bson_free (str);
    }
 }
@@ -566,8 +547,12 @@ bson_strncpy (char *dst,       /* IN */
       return;
    }
 
+/* Prefer strncpy_s for MSVC, or strlcpy, which has additional checks and only
+ * adds one trailing \0 */
 #ifdef _MSC_VER
    strncpy_s (dst, size, src, _TRUNCATE);
+#elif defined(BSON_HAVE_STRLCPY)
+   strlcpy (dst, src, size);
 #else
    strncpy (dst, src, size);
    dst[size - 1] = '\0';
@@ -728,7 +713,7 @@ bson_ascii_strtoll (const char *s, char **e, int base)
 
    c = *tok;
 
-   while (isspace (c)) {
+   while (bson_isspace (c)) {
       c = *++tok;
    }
 
@@ -827,3 +812,10 @@ bson_strcasecmp (const char *s1, const char *s2)
    return strcasecmp (s1, s2);
 #endif
 }
+
+
+bool
+bson_isspace (int c)
+{
+   return c >= -1 && c <= 255 && isspace (c);
+}
