@@ -49,12 +49,14 @@ mongoc_bulk_operation_new (bool ordered)
 {
    mongoc_bulk_operation_t *bulk;
 
-   bulk = (mongoc_bulk_operation_t *) bson_malloc0 (sizeof *bulk);
+   bulk = BSON_ALIGNED_ALLOC0 (mongoc_bulk_operation_t);
    bulk->flags.bypass_document_validation = false;
    bulk->flags.ordered = ordered;
    bulk->server_id = 0;
 
-   _mongoc_array_init (&bulk->commands, sizeof (mongoc_write_command_t));
+   bson_init (&bulk->let);
+
+   mongoc_array_aligned_init (&bulk->commands, mongoc_write_command_t);
    _mongoc_write_result_init (&bulk->result);
 
    return bulk;
@@ -71,8 +73,8 @@ _mongoc_bulk_operation_new (
 {
    mongoc_bulk_operation_t *bulk;
 
-   BSON_ASSERT (client);
-   BSON_ASSERT (collection);
+   BSON_ASSERT_PARAM (client);
+   BSON_ASSERT_PARAM (collection);
 
    bulk = mongoc_bulk_operation_new (flags.ordered);
    bulk->client = client;
@@ -91,10 +93,9 @@ void
 mongoc_bulk_operation_destroy (mongoc_bulk_operation_t *bulk) /* IN */
 {
    mongoc_write_command_t *command;
-   int i;
 
    if (bulk) {
-      for (i = 0; i < bulk->commands.len; i++) {
+      for (size_t i = 0; i < bulk->commands.len; i++) {
          command =
             &_mongoc_array_index (&bulk->commands, mongoc_write_command_t, i);
          _mongoc_write_command_destroy (command);
@@ -102,6 +103,8 @@ mongoc_bulk_operation_destroy (mongoc_bulk_operation_t *bulk) /* IN */
 
       bson_free (bulk->database);
       bson_free (bulk->collection);
+      bson_value_destroy (&bulk->comment);
+      bson_destroy (&bulk->let);
       mongoc_write_concern_destroy (bulk->write_concern);
       _mongoc_array_destroy (&bulk->commands);
 
@@ -135,7 +138,7 @@ mongoc_bulk_operation_destroy (mongoc_bulk_operation_t *bulk) /* IN */
    } while (0)
 
 
-bool
+static bool
 _mongoc_bulk_operation_remove_with_opts (
    mongoc_bulk_operation_t *bulk,
    const bson_t *selector,
@@ -145,14 +148,16 @@ _mongoc_bulk_operation_remove_with_opts (
 {
    mongoc_write_command_t command = {0};
    mongoc_write_command_t *last;
+   bson_t cmd_opts = BSON_INITIALIZER;
    bson_t opts;
    bool has_collation;
    bool ret = false;
+   bool has_delete_hint;
 
    ENTRY;
 
-   BSON_ASSERT (bulk);
-   BSON_ASSERT (selector);
+   BSON_ASSERT_PARAM (bulk);
+   BSON_ASSERT_PARAM (selector);
 
    bson_init (&opts);
 
@@ -174,11 +179,17 @@ _mongoc_bulk_operation_remove_with_opts (
       bson_append_document (&opts, "collation", 9, &remove_opts->collation);
    }
 
+   has_delete_hint = !!(remove_opts->hint.value_type);
+   if (has_delete_hint) {
+      bson_append_value (&opts, "hint", 4, &remove_opts->hint);
+   }
+
    if (bulk->commands.len) {
       last = &_mongoc_array_index (
          &bulk->commands, mongoc_write_command_t, bulk->commands.len - 1);
       if (last->type == MONGOC_WRITE_COMMAND_DELETE) {
          last->flags.has_collation |= has_collation;
+         last->flags.has_delete_hint |= has_delete_hint;
          last->flags.has_multi_write |= (remove_opts->limit == 0);
          _mongoc_write_command_delete_append (last, selector, &opts);
          ret = true;
@@ -186,16 +197,26 @@ _mongoc_bulk_operation_remove_with_opts (
       }
    }
 
+   if (bulk->comment.value_type != BSON_TYPE_EOD) {
+      bson_append_value (&cmd_opts, "comment", 7, &bulk->comment);
+   }
+
+   if (!bson_empty (&bulk->let)) {
+      bson_append_document (&cmd_opts, "let", 3, &bulk->let);
+   }
+
    _mongoc_write_command_init_delete (
-      &command, selector, NULL, &opts, bulk->flags, bulk->operation_id);
+      &command, selector, &cmd_opts, &opts, bulk->flags, bulk->operation_id);
 
    command.flags.has_collation = has_collation;
+   command.flags.has_delete_hint = has_delete_hint;
    command.flags.has_multi_write = (remove_opts->limit == 0);
 
    _mongoc_array_append_val (&bulk->commands, command);
    ret = true;
 
 done:
+   bson_destroy (&cmd_opts);
    bson_destroy (&opts);
    RETURN (ret);
 }
@@ -328,8 +349,8 @@ mongoc_bulk_operation_insert (mongoc_bulk_operation_t *bulk,
 {
    ENTRY;
 
-   BSON_ASSERT (bulk);
-   BSON_ASSERT (document);
+   BSON_ASSERT_PARAM (bulk);
+   BSON_ASSERT_PARAM (document);
 
    if (!mongoc_bulk_operation_insert_with_opts (
           bulk, document, NULL /* opts */, &bulk->result.error)) {
@@ -348,12 +369,13 @@ mongoc_bulk_operation_insert_with_opts (mongoc_bulk_operation_t *bulk,
    mongoc_bulk_insert_opts_t insert_opts;
    mongoc_write_command_t command = {0};
    mongoc_write_command_t *last;
+   bson_t cmd_opts = BSON_INITIALIZER;
    bool ret = false;
 
    ENTRY;
 
-   BSON_ASSERT (bulk);
-   BSON_ASSERT (document);
+   BSON_ASSERT_PARAM (bulk);
+   BSON_ASSERT_PARAM (document);
 
    BULK_RETURN_IF_PRIOR_ERROR;
 
@@ -366,6 +388,9 @@ mongoc_bulk_operation_insert_with_opts (mongoc_bulk_operation_t *bulk,
       GOTO (done);
    }
 
+   /* Note: mongoc_bulk_insert_opts_t specifies allow_extra=False, so there is
+    * no reason to concatenate cmd_opts with &insert_opts.extra. */
+
    if (bulk->commands.len) {
       last = &_mongoc_array_index (
          &bulk->commands, mongoc_write_command_t, bulk->commands.len - 1);
@@ -377,8 +402,12 @@ mongoc_bulk_operation_insert_with_opts (mongoc_bulk_operation_t *bulk,
       }
    }
 
+   if (bulk->comment.value_type != BSON_TYPE_EOD) {
+      bson_append_value (&cmd_opts, "comment", 7, &bulk->comment);
+   }
+
    _mongoc_write_command_init_insert (
-      &command, document, &insert_opts.extra, bulk->flags, bulk->operation_id);
+      &command, document, &cmd_opts, bulk->flags, bulk->operation_id);
 
    _mongoc_array_append_val (&bulk->commands, command);
 
@@ -386,6 +415,8 @@ mongoc_bulk_operation_insert_with_opts (mongoc_bulk_operation_t *bulk,
 
 done:
    _mongoc_bulk_insert_opts_cleanup (&insert_opts);
+   bson_destroy (&cmd_opts);
+
    RETURN (ret);
 }
 
@@ -400,6 +431,7 @@ _mongoc_bulk_operation_update_append (
 {
    mongoc_write_command_t command = {0};
    mongoc_write_command_t *last;
+   bson_t cmd_opts = BSON_INITIALIZER;
    bson_t opts;
    bool has_collation;
    bool has_array_filters;
@@ -432,17 +464,30 @@ _mongoc_bulk_operation_update_append (
       last = &_mongoc_array_index (
          &bulk->commands, mongoc_write_command_t, bulk->commands.len - 1);
       if (last->type == MONGOC_WRITE_COMMAND_UPDATE) {
+         last->flags.has_array_filters |= has_array_filters;
          last->flags.has_collation |= has_collation;
          last->flags.has_update_hint |= has_update_hint;
          last->flags.has_multi_write |= update_opts->multi;
          _mongoc_write_command_update_append (last, selector, document, &opts);
-         bson_destroy (&opts);
-         return;
+         GOTO (done);
       }
    }
 
-   _mongoc_write_command_init_update (
-      &command, selector, document, &opts, bulk->flags, bulk->operation_id);
+   if (bulk->comment.value_type != BSON_TYPE_EOD) {
+      bson_append_value (&cmd_opts, "comment", 7, &bulk->comment);
+   }
+
+   if (!bson_empty (&bulk->let)) {
+      bson_append_document (&cmd_opts, "let", 3, &bulk->let);
+   }
+
+   _mongoc_write_command_init_update (&command,
+                                      selector,
+                                      document,
+                                      &cmd_opts,
+                                      &opts,
+                                      bulk->flags,
+                                      bulk->operation_id);
 
    command.flags.has_array_filters = has_array_filters;
    command.flags.has_collation = has_collation;
@@ -450,6 +495,9 @@ _mongoc_bulk_operation_update_append (
    command.flags.has_multi_write = update_opts->multi;
 
    _mongoc_array_append_val (&bulk->commands, command);
+
+done:
+   bson_destroy (&cmd_opts);
    bson_destroy (&opts);
 }
 
@@ -466,9 +514,9 @@ _mongoc_bulk_operation_update_with_opts (
 {
    ENTRY;
 
-   BSON_ASSERT (bulk);
-   BSON_ASSERT (selector);
-   BSON_ASSERT (document);
+   BSON_ASSERT_PARAM (bulk);
+   BSON_ASSERT_PARAM (selector);
+   BSON_ASSERT_PARAM (document);
 
    if (!_mongoc_validate_update (document, update_opts->validate, error)) {
       RETURN (false);
@@ -633,9 +681,9 @@ mongoc_bulk_operation_replace_one_with_opts (mongoc_bulk_operation_t *bulk,
 
    ENTRY;
 
-   BSON_ASSERT (bulk);
-   BSON_ASSERT (selector);
-   BSON_ASSERT (document);
+   BSON_ASSERT_PARAM (bulk);
+   BSON_ASSERT_PARAM (selector);
+   BSON_ASSERT_PARAM (document);
 
    BULK_RETURN_IF_PRIOR_ERROR;
 
@@ -701,11 +749,10 @@ mongoc_bulk_operation_execute (mongoc_bulk_operation_t *bulk, /* IN */
    mongoc_server_stream_t *server_stream;
    bool ret;
    uint32_t offset = 0;
-   int i;
 
    ENTRY;
 
-   BSON_ASSERT (bulk);
+   BSON_ASSERT_PARAM (bulk);
 
    if (!bulk->client) {
       bson_set_error (error,
@@ -758,7 +805,7 @@ mongoc_bulk_operation_execute (mongoc_bulk_operation_t *bulk, /* IN */
       GOTO (err);
    }
 
-   for (i = 0; i < bulk->commands.len; i++) {
+   for (size_t i = 0u; i < bulk->commands.len; i++) {
       if (bulk->server_id) {
          server_stream =
             mongoc_cluster_stream_for_server (cluster,
@@ -769,7 +816,7 @@ mongoc_bulk_operation_execute (mongoc_bulk_operation_t *bulk, /* IN */
                                               error);
       } else {
          server_stream = mongoc_cluster_stream_for_writes (
-            cluster, bulk->session, reply, error);
+            cluster, bulk->session, NULL, reply, error);
       }
 
       if (!server_stream) {
@@ -827,7 +874,7 @@ void
 mongoc_bulk_operation_set_write_concern (
    mongoc_bulk_operation_t *bulk, const mongoc_write_concern_t *write_concern)
 {
-   BSON_ASSERT (bulk);
+   BSON_ASSERT_PARAM (bulk);
 
    if (bulk->write_concern) {
       mongoc_write_concern_destroy (bulk->write_concern);
@@ -843,7 +890,7 @@ mongoc_bulk_operation_set_write_concern (
 const mongoc_write_concern_t *
 mongoc_bulk_operation_get_write_concern (const mongoc_bulk_operation_t *bulk)
 {
-   BSON_ASSERT (bulk);
+   BSON_ASSERT_PARAM (bulk);
 
    return bulk->write_concern;
 }
@@ -853,7 +900,7 @@ void
 mongoc_bulk_operation_set_database (mongoc_bulk_operation_t *bulk,
                                     const char *database)
 {
-   BSON_ASSERT (bulk);
+   BSON_ASSERT_PARAM (bulk);
 
    if (bulk->database) {
       bson_free (bulk->database);
@@ -867,7 +914,7 @@ void
 mongoc_bulk_operation_set_collection (mongoc_bulk_operation_t *bulk,
                                       const char *collection)
 {
-   BSON_ASSERT (bulk);
+   BSON_ASSERT_PARAM (bulk);
 
    if (bulk->collection) {
       bson_free (bulk->collection);
@@ -880,8 +927,8 @@ mongoc_bulk_operation_set_collection (mongoc_bulk_operation_t *bulk,
 void
 mongoc_bulk_operation_set_client (mongoc_bulk_operation_t *bulk, void *client)
 {
-   BSON_ASSERT (bulk);
-   BSON_ASSERT (client);
+   BSON_ASSERT_PARAM (bulk);
+   BSON_ASSERT_PARAM (client);
 
    if (bulk->session) {
       BSON_ASSERT (bulk->session->client == client);
@@ -903,8 +950,8 @@ mongoc_bulk_operation_set_client_session (
    mongoc_bulk_operation_t *bulk,
    struct _mongoc_client_session_t *client_session)
 {
-   BSON_ASSERT (bulk);
-   BSON_ASSERT (client_session);
+   BSON_ASSERT_PARAM (bulk);
+   BSON_ASSERT_PARAM (client_session);
 
    if (bulk->client) {
       BSON_ASSERT (bulk->client == client_session->client);
@@ -917,7 +964,7 @@ mongoc_bulk_operation_set_client_session (
 uint32_t
 mongoc_bulk_operation_get_hint (const mongoc_bulk_operation_t *bulk)
 {
-   BSON_ASSERT (bulk);
+   BSON_ASSERT_PARAM (bulk);
 
    return bulk->server_id;
 }
@@ -927,7 +974,7 @@ void
 mongoc_bulk_operation_set_hint (mongoc_bulk_operation_t *bulk,
                                 uint32_t server_id)
 {
-   BSON_ASSERT (bulk);
+   BSON_ASSERT_PARAM (bulk);
 
    bulk->server_id = server_id;
 }
@@ -937,7 +984,43 @@ void
 mongoc_bulk_operation_set_bypass_document_validation (
    mongoc_bulk_operation_t *bulk, bool bypass)
 {
-   BSON_ASSERT (bulk);
+   BSON_ASSERT_PARAM (bulk);
 
    bulk->flags.bypass_document_validation = bypass;
 }
+
+
+void
+mongoc_bulk_operation_set_comment (mongoc_bulk_operation_t *bulk,
+                                   const bson_value_t *comment)
+{
+   BSON_ASSERT_PARAM (bulk);
+   BSON_ASSERT_PARAM (comment);
+   BSON_ASSERT (comment->value_type != BSON_TYPE_EOD);
+
+   /* This method cannot be called after appending operations, as the CRUD spec
+    * states the option should apply to all commands. Since commands are
+    * initialized as operations are added, allowing "comment" to be changed at
+    * any time could violate that contract. */
+   BSON_ASSERT (bulk->commands.len == 0);
+
+   bson_value_destroy (&bulk->comment);
+   bson_value_copy (comment, &bulk->comment);
+}
+
+
+void
+mongoc_bulk_operation_set_let (mongoc_bulk_operation_t *bulk, const bson_t *let)
+{
+   BSON_ASSERT_PARAM (bulk);
+   BSON_ASSERT_PARAM (let);
+
+   /* This method cannot be called after appending operations, as the CRUD spec
+    * states the option should apply to all commands (excluding insert). Since
+    * commands are initialized as operations are added, allowing "let" to be
+    * changed at any time could violate that contract. */
+   BSON_ASSERT (bulk->commands.len == 0);
+
+   bson_destroy (&bulk->let);
+   bson_copy_to (let, &bulk->let);
+}
