@@ -14,19 +14,22 @@
  * limitations under the License.
  */
 
+#include "mongoc-prelude.h"
+
 
 #ifndef MONGOC_BULK_OPERATION_H
 #define MONGOC_BULK_OPERATION_H
 
 
-#include <bson.h>
+#include <bson/bson.h>
 
 #include "mongoc-macros.h"
 #include "mongoc-write-concern.h"
 
-#define MONGOC_BULK_WRITE_FLAGS_INIT                     \
-   {                                                     \
-      true, MONGOC_BYPASS_DOCUMENT_VALIDATION_DEFAULT, 0 \
+/* ordered, bypass_document_validation, has_collation, multi */
+#define MONGOC_BULK_WRITE_FLAGS_INIT \
+   {                                 \
+      true, false, 0                 \
    }
 
 BSON_BEGIN_DECLS
@@ -112,6 +115,12 @@ mongoc_bulk_operation_update_one_with_opts (mongoc_bulk_operation_t *bulk,
 MONGOC_EXPORT (void)
 mongoc_bulk_operation_set_bypass_document_validation (
    mongoc_bulk_operation_t *bulk, bool bypass);
+MONGOC_EXPORT (void)
+mongoc_bulk_operation_set_comment (mongoc_bulk_operation_t *bulk,
+                                   const bson_value_t *comment);
+MONGOC_EXPORT (void)
+mongoc_bulk_operation_set_let (mongoc_bulk_operation_t *bulk,
+                               const bson_t *let);
 
 
 /*
@@ -120,7 +129,7 @@ mongoc_bulk_operation_set_bypass_document_validation (
  * collections.
  */
 MONGOC_EXPORT (mongoc_bulk_operation_t *)
-mongoc_bulk_operation_new (bool ordered);
+mongoc_bulk_operation_new (bool ordered) BSON_GNUC_WARN_UNUSED_RESULT;
 MONGOC_EXPORT (void)
 mongoc_bulk_operation_set_write_concern (
    mongoc_bulk_operation_t *bulk, const mongoc_write_concern_t *write_concern);
