@@ -107,7 +107,7 @@ _make_command (mongoc_change_stream_t *stream, bson_t *command)
    bson_iter_t iter;
    bson_t change_stream_stage; /* { $changeStream: <change_stream_doc> } */
    bson_t change_stream_doc;
-   bson_t pipeline;
+   bson_array_builder_t *pipeline;
    bson_t cursor_doc;
 
    if (stream->change_stream_type == MONGOC_CHANGE_STREAM_COLLECTION) {
@@ -117,13 +117,23 @@ _make_command (mongoc_change_stream_t *stream, bson_t *command)
       bson_append_int32 (command, "aggregate", 9, 1);
    }
 
-   bson_append_array_begin (command, "pipeline", 8, &pipeline);
+   bson_append_array_builder_begin (command, "pipeline", 8, &pipeline);
 
    /* append the $changeStream stage. */
-   bson_append_document_begin (&pipeline, "0", 1, &change_stream_stage);
+   bson_array_builder_append_document_begin (pipeline, &change_stream_stage);
    bson_append_document_begin (
       &change_stream_stage, "$changeStream", 13, &change_stream_doc);
-   bson_concat (&change_stream_doc, stream->full_document);
+   if (stream->full_document) {
+      bson_concat (&change_stream_doc, stream->full_document);
+   }
+   if (stream->full_document_before_change) {
+      bson_concat (&change_stream_doc, stream->full_document_before_change);
+   }
+   if (stream->show_expanded_events) {
+      BSON_APPEND_BOOL (&change_stream_doc,
+                        "showExpandedEvents",
+                        stream->show_expanded_events);
+   }
 
    if (stream->resumed) {
       /* Change stream spec: Resume Process */
@@ -182,29 +192,23 @@ _make_command (mongoc_change_stream_t *stream, bson_t *command)
       bson_append_bool (&change_stream_doc, "allChangesForCluster", 20, true);
    }
    bson_append_document_end (&change_stream_stage, &change_stream_doc);
-   bson_append_document_end (&pipeline, &change_stream_stage);
+   bson_array_builder_append_document_end (pipeline, &change_stream_stage);
 
    /* Append user pipeline if it exists */
    if (bson_iter_init_find (&iter, &stream->pipeline_to_append, "pipeline") &&
        BSON_ITER_HOLDS_ARRAY (&iter)) {
       bson_iter_t child_iter;
-      uint32_t key_int = 1;
-      char buf[16];
-      const char *key_str;
 
       BSON_ASSERT (bson_iter_recurse (&iter, &child_iter));
       while (bson_iter_next (&child_iter)) {
          /* the user pipeline may consist of invalid stages or non-documents.
           * append anyway, and rely on the server error. */
-         size_t keyLen =
-            bson_uint32_to_string (key_int, &key_str, buf, sizeof (buf));
-         bson_append_value (
-            &pipeline, key_str, (int) keyLen, bson_iter_value (&child_iter));
-         ++key_int;
+         bson_array_builder_append_value (pipeline,
+                                          bson_iter_value (&child_iter));
       }
    }
 
-   bson_append_array_end (command, &pipeline);
+   bson_append_array_builder_end (command, pipeline);
 
    /* Add batch size if needed */
    bson_append_document_begin (command, "cursor", 6, &cursor_doc);
@@ -243,6 +247,11 @@ _make_cursor (mongoc_change_stream_t *stream)
    bson_init (&command);
    bson_copy_to (&(stream->opts.extra), &command_opts);
 
+   if (stream->opts.comment.value_type != BSON_TYPE_EOD) {
+      bson_append_value (&command_opts, "comment", 7, &stream->opts.comment);
+      bson_append_value (&getmore_opts, "comment", 7, &stream->opts.comment);
+   }
+
    if (bson_iter_init_find (&iter, &command_opts, "sessionId")) {
       if (!_mongoc_client_session_from_iter (
              stream->client, &iter, &cs, &stream->err)) {
@@ -279,13 +288,12 @@ _make_cursor (mongoc_change_stream_t *stream)
       goto cleanup;
    }
 
-   server_stream =
-      mongoc_cluster_stream_for_reads (&stream->client->cluster,
-                                       stream->read_prefs,
-                                       cs,
-                                       &reply,
-                                       /* Not aggregate-with-write */ false,
-                                       &stream->err);
+   server_stream = mongoc_cluster_stream_for_reads (&stream->client->cluster,
+                                                    stream->read_prefs,
+                                                    cs,
+                                                    NULL,
+                                                    &reply,
+                                                    &stream->err);
    if (!server_stream) {
       bson_destroy (&stream->err_doc);
       bson_copy_to (&reply, &stream->err_doc);
@@ -408,13 +416,22 @@ _change_stream_init (mongoc_change_stream_t *stream,
       return;
    }
 
-   stream->full_document = BCON_NEW ("fullDocument", stream->opts.fullDocument);
+   if (stream->opts.fullDocument) {
+      stream->full_document =
+         BCON_NEW ("fullDocument", stream->opts.fullDocument);
+   }
+
+   if (stream->opts.fullDocumentBeforeChange) {
+      stream->full_document_before_change = BCON_NEW (
+         "fullDocumentBeforeChange", stream->opts.fullDocumentBeforeChange);
+   }
 
    _mongoc_timestamp_set (&stream->operation_time,
                           &stream->opts.startAtOperationTime);
 
    stream->batch_size = stream->opts.batchSize;
    stream->max_await_time_ms = stream->opts.maxAwaitTimeMS;
+   stream->show_expanded_events = stream->opts.showExpandedEvents;
 
    /* Accept two forms of user pipeline:
     * 1. A document like: { "pipeline": [...] }
@@ -452,8 +469,7 @@ _mongoc_change_stream_new_from_collection (const mongoc_collection_t *coll,
    mongoc_change_stream_t *stream;
    BSON_ASSERT (coll);
 
-   stream =
-      (mongoc_change_stream_t *) bson_malloc0 (sizeof (mongoc_change_stream_t));
+   stream = BSON_ALIGNED_ALLOC0 (mongoc_change_stream_t);
    stream->db = bson_strdup (coll->db);
    stream->coll = bson_strdup (coll->collection);
    stream->read_prefs = mongoc_read_prefs_copy (coll->read_prefs);
@@ -472,8 +488,7 @@ _mongoc_change_stream_new_from_database (const mongoc_database_t *db,
    mongoc_change_stream_t *stream;
    BSON_ASSERT (db);
 
-   stream =
-      (mongoc_change_stream_t *) bson_malloc0 (sizeof (mongoc_change_stream_t));
+   stream = BSON_ALIGNED_ALLOC0 (mongoc_change_stream_t);
    stream->db = bson_strdup (db->name);
    stream->coll = NULL;
    stream->read_prefs = mongoc_read_prefs_copy (db->read_prefs);
@@ -492,8 +507,7 @@ _mongoc_change_stream_new_from_client (mongoc_client_t *client,
    mongoc_change_stream_t *stream;
    BSON_ASSERT (client);
 
-   stream =
-      (mongoc_change_stream_t *) bson_malloc0 (sizeof (mongoc_change_stream_t));
+   stream = BSON_ALIGNED_ALLOC0 (mongoc_change_stream_t);
    stream->db = bson_strdup ("admin");
    stream->coll = NULL;
    stream->read_prefs = mongoc_read_prefs_copy (client->read_prefs);
@@ -656,6 +670,7 @@ mongoc_change_stream_destroy (mongoc_change_stream_t *stream)
    bson_destroy (&stream->pipeline_to_append);
    bson_destroy (&stream->resume_token);
    bson_destroy (stream->full_document);
+   bson_destroy (stream->full_document_before_change);
    bson_destroy (&stream->err_doc);
    _mongoc_change_stream_opts_cleanup (&stream->opts);
    mongoc_cursor_destroy (stream->cursor);
